/**
 * Copyright (c) 2014-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.registry_overlay.core;

import java.util.Map.Entry;
import java.util.concurrent.TimeUnit;
import java.util.logging.Logger;

import org.ow2.petals.registry_overlay.api.DataModelApi;

import com.hazelcast.core.Client;
import com.hazelcast.core.HazelcastInstance;
import com.hazelcast.core.ILock;
import com.hazelcast.core.IMap;

/**
 * A client listener cleaning topologies on client disconnection. When no more Petals ESB container of a topology is
 * running, its data registry must be cleaned.
 * 
 * TODO shouldn't it be when all the member of the topology left it? and not the client themselves...
 * 
 * @author Christophe DENEUX - Linagora
 * 
 */
public class ClientListener implements com.hazelcast.core.ClientListener {

    private final Logger log;

    private final HazelcastInstance hzlInstance;

    private final long clientMapLockWaitTime;

    public ClientListener(final Logger log, final HazelcastInstance hzlInstance, final long clientMapLockWaitTime) {
        this.log = log;
        this.hzlInstance = hzlInstance;
        this.clientMapLockWaitTime = clientMapLockWaitTime;
    }

    @Override
    public void clientConnected(final Client client) {
        // NOP: We are interested only by client disconnections to clean topologies
    }

    @Override
    public void clientDisconnected(final Client client) {
        this.log.fine("A client was disconnected");
        final ILock clientsMapLock = DataModelApi.getClientsMapLock(this.hzlInstance);
        final IMap<String, String> clientsMap = DataModelApi.getClientsMap(this.hzlInstance);
        try {
            if (clientsMapLock.tryLock(this.clientMapLockWaitTime, TimeUnit.SECONDS)) {
                try {
                    final String clientRegistryName = clientsMap.get(client.getUuid());
                    if (clientRegistryName != null) {
                        // Check if another container of the current topology is always running
                        boolean anotherContainerRunning = false;
                        for (final Entry<String, String> entry : clientsMap.entrySet()) {
                            if (entry.getValue().equals(clientRegistryName) && !entry.getKey().equals(client.getUuid())) {
                                anotherContainerRunning = true;
                                break;
                            }
                        }
                        if (!anotherContainerRunning) {
                            if (!DataModelApi.cleanTopology(this.hzlInstance, clientRegistryName)) {
                                // This should not occur because the topology is registered
                                this.log.warning("The registry to delete ('" + clientRegistryName
                                        + "') does not more exist");
                            }
                        }
                        DataModelApi.getClientsMap(this.hzlInstance).remove(client.getUuid());
                    } else {
                        this.log.warning("The client disconnected is not referenced in the clients map");
                    }
                } finally {
                    clientsMapLock.unlock();
                }
            } else {
                this.log.warning("Unable to lock the clients map, the registry can be inconsistent after topology restart");
            }
        } catch (final InterruptedException e) {
            this.log.warning("The clients map lock acquisition was interrupted, the registry can be inconsistent after topology restart");
        }
    }

}
