/**
 * Copyright (c) 2013-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.registry_overlay.core;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.lang.management.ManagementFactory;
import java.net.InetAddress;
import java.net.MalformedURLException;
import java.net.UnknownHostException;
import java.rmi.RemoteException;
import java.rmi.registry.LocateRegistry;
import java.rmi.registry.Registry;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.management.InstanceAlreadyExistsException;
import javax.management.InstanceNotFoundException;
import javax.management.MBeanRegistrationException;
import javax.management.MBeanServer;
import javax.management.MalformedObjectNameException;
import javax.management.NotCompliantMBeanException;
import javax.management.ObjectName;
import javax.management.remote.JMXConnectorServer;
import javax.management.remote.JMXConnectorServerFactory;
import javax.management.remote.JMXServiceURL;

import org.ow2.petals.registry_overlay.api.Constants;
import org.ow2.petals.registry_overlay.api.DataModelApi;
import org.ow2.petals.registry_overlay.api.RegistryEndpointFactory;
import org.ow2.petals.registry_overlay.configuration.generated.Credentials;
import org.ow2.petals.registry_overlay.configuration.generated.ManagementConsole;
import org.ow2.petals.registry_overlay.configuration.generated.Member;
import org.ow2.petals.registry_overlay.configuration.generated.Members;
import org.ow2.petals.registry_overlay.configuration.generated.PetalsRegistryOverlay;
import org.ow2.petals.registry_overlay.core.exception.ConfigurationException;
import org.ow2.petals.registry_overlay.core.exception.InvalidJmxPortException;
import org.ow2.petals.registry_overlay.core.exception.MemberIdentifierMissingException;
import org.ow2.petals.registry_overlay.core.exception.MissingCredentialsException;
import org.ow2.petals.registry_overlay.core.exception.MissingCredentialsGroupException;
import org.ow2.petals.registry_overlay.core.exception.MissingCredentialsPasswordException;
import org.ow2.petals.registry_overlay.core.exception.MissingMemberIdException;
import org.ow2.petals.registry_overlay.core.exception.MissingMembersException;
import org.ow2.petals.registry_overlay.core.exception.NotAClusterMemberException;
import org.ow2.petals.registry_overlay.core.exception.SeveralMemberIdException;
import org.ow2.petals.registry_overlay.core.mbean.Admin;
import org.ow2.petals.registry_overlay.core.mbean.AdminMBean;

import com.hazelcast.config.Config;
import com.hazelcast.config.GroupConfig;
import com.hazelcast.config.JoinConfig;
import com.hazelcast.config.ManagementCenterConfig;
import com.hazelcast.config.MapConfig;
import com.hazelcast.config.MapIndexConfig;
import com.hazelcast.config.NearCacheConfig;
import com.hazelcast.config.NetworkConfig;
import com.hazelcast.config.SerializationConfig;
import com.hazelcast.config.TcpIpConfig;
import com.hazelcast.core.ClientService;
import com.hazelcast.core.Hazelcast;
import com.hazelcast.core.HazelcastInstance;

/**
 * A node of the registry in overlay.
 * 
 * @author Christophe DENEUX - Linagora
 * 
 */
public class RegistryOverlayNode {

    /**
     * Logger name
     */
    public static final String LOGGER_NAME = "Petals.Registry.Overlay";

    /**
     * Default member hostname
     */
    public static final String DEFAULT_HOST = "localhost";

    /**
     * Default Hazelcast management center URL
     */
    public static final String DEFAULT_MNGT_CENTER_URL = "http://localhost:8080/mancenter";

    /**
     * Default wait time, in second, to acquire the lock of the client map
     */
    public static final long DEFAULT_CLIENTSMAP_LOCK_WAITTIME = 10;

    /**
     * The Hazelcast configuration of the node used as base configuration for each instance of Hazelcast.
     */
    private final Config hzlBaseCfg;

    /**
     * The hazelcast instance
     */
    private HazelcastInstance instance;

    /**
     * The JMX connector
     */
    private JMXConnectorServer jmxConnectorServer = null;

    /**
     * A lock to manage the Hazelcast instances;
     */
    private final Object instancesLock = new Object();

    private final Logger log = Logger.getLogger(RegistryOverlayNode.LOGGER_NAME);

    /**
     * The administration MBean.
     */
    private final AdminMBean adminMBean;

    /**
     * The name of the administration MBean
     */
    private ObjectName adminMBeanName = null;

    /**
     * Flag indicating the state of the Petals Registry Overlay node. <code>true</code> if the node is starting,
     * started; <code>false</code> if the node is stopping or stopped.
     */
    private volatile boolean isRunning = false;

    /**
     * The max wait time to acquire a lock on the clients map
     */
    private long clientsMapLockWaitTime;

    /**
     * Registration identifier of the client listener
     */
    private String clientListenerId = null;

    /**
     * Create a node of the Petals Registry Overlay
     * 
     * @param petalsRegistryOverlay
     *            The Petals Registry Overlay configuration
     * @param memberProperties
     *            Configuration properties of the current member
     * @throws ConfigurationException
     *             An error about the Petals Registry Overlay configuration occurs
     */
    public RegistryOverlayNode(final PetalsRegistryOverlay petalsRegistryOverlay,
            final Properties memberProperties)
            throws ConfigurationException {

        final String memberIdToCreate = memberProperties.getProperty(Constants.MEMBER_PROPERTY_NAME_MEMBER_ID);
        if (memberIdToCreate == null || memberIdToCreate.trim().isEmpty()) {
            throw new MemberIdentifierMissingException();
        }

        // =====================================================================
        // == Create the Hazelcast configuration of the current cluster member
        // =====================================================================
        this.hzlBaseCfg = new Config();

        // Enable Hazelcast JMX MBean
        this.hzlBaseCfg.setProperty("hazelcast.jmx", Boolean.TRUE.toString());

        final Credentials credentials = petalsRegistryOverlay.getCredentials();
        if (credentials == null) {
            throw new MissingCredentialsException();
        }
        final String group = credentials.getGroup();
        if (group == null || group.isEmpty()) {
            throw new MissingCredentialsGroupException();
        }
        final String pwd = credentials.getPassword();
        if (pwd == null || pwd.isEmpty()) {
            throw new MissingCredentialsPasswordException();
        }
        final GroupConfig groupConfig = new GroupConfig(group, pwd);
        this.hzlBaseCfg.setGroupConfig(groupConfig);

        final ManagementConsole mngtConsole = petalsRegistryOverlay.getManagementConsole();
        if (mngtConsole == null || mngtConsole.isEnable()) {
            final ManagementCenterConfig hzlMngtCenterConfig = new ManagementCenterConfig();
            hzlMngtCenterConfig.setEnabled(true);
            if (mngtConsole == null || mngtConsole.getValue() == null || mngtConsole.getValue().trim().isEmpty()) {
                hzlMngtCenterConfig.setUrl(DEFAULT_MNGT_CENTER_URL);
            } else {
                hzlMngtCenterConfig.setUrl(mngtConsole.getValue());
            }
            this.hzlBaseCfg.setManagementCenterConfig(hzlMngtCenterConfig);
        }

        final NetworkConfig hzlNetworkCfg = this.hzlBaseCfg.getNetworkConfig();
        // Ports are explicitly defined
        hzlNetworkCfg.setPortAutoIncrement(false);
        hzlNetworkCfg.setReuseAddress(true);

        // A TCP/IP cluster is required
        final JoinConfig hzlNetworkJoinCfg = hzlNetworkCfg.getJoin();
        hzlNetworkJoinCfg.getAwsConfig().setEnabled(false);
        hzlNetworkJoinCfg.getMulticastConfig().setEnabled(false);
        final TcpIpConfig hzlTcpIpCfg = hzlNetworkJoinCfg.getTcpIpConfig();
        hzlTcpIpCfg.setEnabled(true);

        final Members members = petalsRegistryOverlay.getMembers();
        if (members == null) {
            throw new MissingMembersException();
        }
        final List<Member> memberList = members.getMember();
        if (memberList == null || memberList.isEmpty()) {
            throw new MissingMembersException();
        }
        String memberHost = null;
        for (final Member member : memberList) {
            final String memberId = member.getId();
            final int port = member.getPort() == null ? Constants.DEFAULT_PORT : member.getPort().intValue();
            final String host = member.getValue() == null ? DEFAULT_HOST : member.getValue();
            if (memberId == null || memberId.isEmpty()) {
                throw new MissingMemberIdException();
            } else if (member.getId().equals(memberIdToCreate)) {
                // this one is me
                if (memberHost != null) {
                    throw new SeveralMemberIdException(memberIdToCreate);
                }
                // TODO: Add a check about the host (should match the listen host)
                memberHost = host;
                // TODO: Add a check about the port value
                hzlNetworkCfg.setPort(port);
            } else {
                // this one is another member
                this.log.config("Adding another member: " + host + ":" + port);
                hzlTcpIpCfg.addMember(host + ":" + port);
            }
        }

        if (memberHost == null) {
            throw new NotAClusterMemberException(memberIdToCreate);
        }

        // if the listen interface is defined in the conf we use it, if not we use the default value
        String listenHost = memberProperties.getProperty(Constants.MEMBER_PROPERTY_NAME_LISTEN_HOST);
        if (listenHost == null || listenHost.trim().isEmpty()) {
            this.log.config("No value for " + Constants.MEMBER_PROPERTY_NAME_LISTEN_HOST
                    + ", member host from cluster definition used: '" + memberHost + "'");
            // let's use the default value
            listenHost = memberHost;
        }
        // now let's transform this to something understood by hazelcast
        if (listenHost.trim().equals("0.0.0.0")) {
            // let's disable interfaces (it will use 0.0.0.0)
            listenHost = null;
            this.log.config("Listen interface is set to 0.0.0.0, disabling specific interfaces listening in Hazelcast");
        } else {
            try {
                final InetAddress resolvedAddress = InetAddress.getByName(listenHost);
                final String previousListenHost = listenHost;
                listenHost = resolvedAddress.getHostAddress();
                if (previousListenHost != listenHost) {
                    this.log.config("Listen interface is set to " + previousListenHost + " and was resolved to the ip "
                            + listenHost);
                }
            } catch (final UnknownHostException e) {
                throw new ConfigurationException("Can't resolve the value of "
                        + Constants.MEMBER_PROPERTY_NAME_LISTEN_HOST + ": '" + listenHost + "'");
            }
        }

        if (listenHost != null) {
            this.log.config("Listening on " + listenHost + ":" + hzlNetworkCfg.getPort());
            this.hzlBaseCfg.setProperty("hazelcast.socket.bind.any", "false");
            hzlNetworkCfg.getInterfaces().setEnabled(true).addInterface(listenHost);
        }

        // We use a wildcarded name to match all maps
        // TODO should the near cache be configured between members or simply for clients?!
        // I think only clients!
        this.hzlBaseCfg.addMapConfig(new MapConfig("*").setNearCacheConfig(new NearCacheConfig()));
        
        final MapConfig endpointsMapConfig = new MapConfig(DataModelApi.getEndpointsMapName("*"));
        this.hzlBaseCfg.addMapConfig(endpointsMapConfig);
        endpointsMapConfig.addMapIndexConfig(new MapIndexConfig("interfaces[any]", false));
        endpointsMapConfig.addMapIndexConfig(new MapIndexConfig("serviceName", false));
        endpointsMapConfig.addMapIndexConfig(new MapIndexConfig("type", false));

        this.hzlBaseCfg.setSerializationConfig(new SerializationConfig().addDataSerializableFactoryClass(
                RegistryEndpointFactory.REGISTRY_ENDPOINT_FACTORY_ID, RegistryEndpointFactory.class));
        
        // ==================================================
        // == JMX configuration of the current cluster member
        // ==================================================
        final String enableJmxConnector = memberProperties
                .getProperty(Constants.MEMBER_PROPERTY_NAME_ENABLE_JMX_CONNECTOR);
        if (enableJmxConnector == null || enableJmxConnector.trim().isEmpty()
                || Boolean.parseBoolean(enableJmxConnector)) {
            this.createJmxConnector(memberProperties, memberHost);
        } else {
            this.log.info("JMX Connector of the Petals Registry Overlay is disable");
        }

        this.adminMBean = new Admin(this);
        try {
            this.adminMBeanName = new ObjectName("PetalsRegistryOverlay-" + memberIdToCreate + ":name="
                    + AdminMBean.MBEAN_NAME + ",type=service");
        } catch (final MalformedObjectNameException e) {
            // This error should not occur if all checks are correctly realized.
            this.log.log(Level.WARNING, "Unable to register the administration MBean.", e);
        }

        // ====================================================
        // == Other configuration of the current cluster member
        // ====================================================
        final String clientsMapLockWaitTimeStr = memberProperties
                .getProperty(Constants.MEMBER_PROPERTY_NAME_CLIENTSMAP_LOCK_WAITTIME);
        if (clientsMapLockWaitTimeStr != null && !clientsMapLockWaitTimeStr.trim().isEmpty()) {
            try {
                this.clientsMapLockWaitTime = Long.parseLong(clientsMapLockWaitTimeStr);
            } catch (final NumberFormatException e) {
                this.log.warning(String.format("Invalid value for property '%s': '%s'. Default value used.",
                        Constants.MEMBER_PROPERTY_NAME_CLIENTSMAP_LOCK_WAITTIME, clientsMapLockWaitTimeStr));
                this.clientsMapLockWaitTime = DEFAULT_CLIENTSMAP_LOCK_WAITTIME;
            }
        } else {
            this.clientsMapLockWaitTime = DEFAULT_CLIENTSMAP_LOCK_WAITTIME;
        }
        this.log.config("Client map lock max wait time, in seconds: " + this.clientsMapLockWaitTime);

    }

    /**
     * Create the JMX connector
     * 
     * @param memberProperties
     *            Configuration properties of the current member
     * @param memberHost
     *            Hostname or IP address of the current member
     * @throws ConfigurationException
     *             An error occurs about the configuration
     */
    private void createJmxConnector(final Properties memberProperties, final String memberHost)
            throws ConfigurationException {
        // Retrieve the port and host
        final int jmxPort = RegistryOverlayNode.getJmxPort(memberProperties);

        // create an RMI connector
        try {
            final Registry registry = LocateRegistry.getRegistry(jmxPort);
            try {
                registry.list();
            } catch (final Exception e) {
                LocateRegistry.createRegistry(jmxPort);
            }

            // Note: the host is simply here for information, because it is the RMI Registry created above listening and
            // it always listen to ALL interfaces on the provided port
            final JMXServiceURL jmxUrl = new JMXServiceURL(getJmxServiceURL(memberHost, jmxPort));
            final Map<String, Object> env = this.setupJMXSecurity(memberProperties);
            this.jmxConnectorServer = JMXConnectorServerFactory.newJMXConnectorServer(jmxUrl, env,
                    ManagementFactory.getPlatformMBeanServer());
        } catch (final RemoteException e) {
            this.log.log(Level.WARNING, "Failed to start JMX RMI connector service", e);
        } catch (final MalformedURLException e) {
            this.log.log(Level.WARNING, "Failed to start JMX RMI connector service", e);
        } catch (final IOException e) {
            this.log.log(Level.WARNING, "Failed to start JMX RMI connector service", e);
        }

    }

    /**
     * Retrieve the JMX port from the configuration of the node
     * 
     * @param memberProperties
     *            Configuration properties of the current member
     * @return The JMX port of the current node
     * @throws ConfigurationException
     *             An error occurs about the configuration
     */
    public static final int getJmxPort(final Properties memberProperties) throws ConfigurationException {
        final String jmxPortStr = memberProperties.getProperty(Constants.MEMBER_PROPERTY_NAME_JMX_PORT);
        final int jmxPort;
        if (jmxPortStr != null && !jmxPortStr.trim().isEmpty()) {
            try {
                jmxPort = Integer.parseInt(jmxPortStr);
            } catch (final NumberFormatException e1) {
                throw new InvalidJmxPortException(jmxPortStr);
            }
        } else {
            jmxPort = Constants.MEMBER_PROPERTY_DEFAULT_JMX_PORT;
        }
        return jmxPort;
    }

    /**
     * Build the JMX Service URL.
     * 
     * @param memberHost
     *            the member hostname
     * @param jmxPort
     *            the JMX port
     * @return the URL in String form
     */
    public static final String getJmxServiceURL(final String memberHost, final int jmxPort) {
        // The RMI server definition
        final StringBuffer rmiServer = new StringBuffer("rmi://");
        rmiServer.append(memberHost).append(":").append(jmxPort);
        // The JMX Service over RMI URL
        final StringBuffer jmxServiceURL = new StringBuffer("service:jmx:");
        jmxServiceURL.append(rmiServer.toString()).append("/jndi/").append(rmiServer.toString())
                .append("/jmxRmiConnector");

        return jmxServiceURL.toString();
    }

    /**
     * <p>
     * Setup JMX security:
     * <ul>
     * <li>create a $PETALS_HOME/work/jmx/password.properties file containing jmx user/pwd values</li>
     * <li>create a $PETALS_HOME/work/jmx/access.properties file containing jmx user and readwrite</li>
     * </ul>
     * </p>
     * return a hashmap referencing these files
     * 
     * @param memberProperties
     *            Configuration properties of the current member
     * @return the env hashmap
     */
    private Map<String, Object> setupJMXSecurity(final Properties memberProperties) {
        
        String username = memberProperties.getProperty(Constants.MEMBER_PROPERTY_NAME_JMX_USER);
        if (username == null || username.trim().isEmpty()) {
            this.log.warning("No JMX user set. The default one will be used.");
            username = Constants.MEMBER_PROPERTY_DEFAULT_JMX_USER;
        }
        
        String password = memberProperties.getProperty(Constants.MEMBER_PROPERTY_NAME_JMX_PWD);
        if (password == null || password.trim().isEmpty()) {
            this.log.warning("No JMX password set. The default one will be used.");
            password = Constants.MEMBER_PROPERTY_DEFAULT_JMX_PWD;
        }

        final Properties pwdProperties = new Properties();
        pwdProperties.put(username, password);

        final Properties accessProperties = new Properties();
        accessProperties.put(username, "readwrite");

        final Map<String, Object> env = new HashMap<String, Object>();
        try {
            final File passwordFile = File.createTempFile("password.", ".properties");
            try {
                pwdProperties.store(new FileOutputStream(passwordFile), "role and password for JMX remote connection");
                env.put("jmx.remote.x.password.file", passwordFile.getAbsolutePath());
            } catch (final IOException e) {
                this.log.warning("Can not create " + passwordFile.getAbsolutePath() + " file to store JMX password");
            }
        } catch (final IOException e) {
            this.log.warning("Can not create the file to store JMX password");
        }

        try {
            final File accessFile = File.createTempFile("access.", ".properties");
            try {
                accessProperties.store(new FileOutputStream(accessFile), "role and right for JMX remote connection");
                env.put("jmx.remote.x.access.file", accessFile.getAbsolutePath());
            } catch (final IOException e) {
                this.log.warning("Can not create " + accessFile.getAbsolutePath() + " file to store JMX rights");
            }
        } catch (IOException e) {
            this.log.warning("Can not create the file to store JMX rights");
        }

        return env;
    }

    /**
     * Starts the node
     */
    public void start() {

        this.isRunning = true;

        synchronized (this.instancesLock) {
            if (this.instance != null) {
                // The instance was created on a previous startup, next stopped and now the instance must be restarted
                this.log.fine("Restarting a instance that was created on a previous startup and stopped");
                this.newHzlInstance();
            } else {
                this.log.fine("First startup");
                this.newHzlInstance();
            }
        }

        try {
            if (this.jmxConnectorServer != null) {
                this.jmxConnectorServer.start();
                this.log.info("JMX server started at : " + "service:jmx:rmi://"
                        + this.jmxConnectorServer.getAddress().getURLPath());
            }
        } catch (final IOException e) {
            this.log.log(Level.WARNING, "An error occurs during startup of the JMX connector server", e);
        }

        // MBean registration
        final MBeanServer mbeanServer = ManagementFactory.getPlatformMBeanServer();
        try {
            mbeanServer.registerMBean(this.adminMBean, this.adminMBeanName);
        } catch (final InstanceAlreadyExistsException e) {
            this.log.warning("The administration MBean is already registered.");
        } catch (final MBeanRegistrationException e) {
            this.log.log(Level.WARNING, "An error occurs registering the administration MBean", e);
        } catch (final NotCompliantMBeanException e) {
            // A bug occurs
            this.log.log(Level.SEVERE, "A bug occurs.", e);
        }
    }

    /**
     * Create and start a new Hazlcast instance
     */
    private void newHzlInstance() {
        this.instance = Hazelcast.newHazelcastInstance(this.hzlBaseCfg);

        final ClientService hzlClientService = this.instance.getClientService();
        this.clientListenerId = hzlClientService.addClientListener(new ClientListener(this.log, this.instance,
                this.clientsMapLockWaitTime));
    }

    /**
     * Stops the node
     */
    public void stop() {

        this.isRunning = false;
        this.log.info("Stopping Petals Registry Overlay local member...");

        try {
            if (this.jmxConnectorServer != null) {
                this.jmxConnectorServer.stop();
            }
        } catch (final IOException e) {
            this.log.log(Level.WARNING, "An error occurs during startup of the JMX connector server", e);
        }

        synchronized (this.instancesLock) {
            if (this.instance == null) {
                this.log.log(Level.SEVERE, "No Hazelcast instance to stop.");
            } else {
                if (!this.instance.getClientService().removeClientListener(this.clientListenerId)) {
                    this.log.warning("Unable to remove the client listener");
                }
                this.instance.getLifecycleService().shutdown();
                // TODO: add a timeout before to force the stop
            }
        }

        // MBean deregistration
        final MBeanServer mbeanServer = ManagementFactory.getPlatformMBeanServer();
        try {
            mbeanServer.unregisterMBean(this.adminMBeanName);
        } catch (final InstanceNotFoundException e) {
            this.log.warning("The administration MBean was not found for deregistration.");
        } catch (final MBeanRegistrationException e) {
            this.log.log(Level.WARNING, "An error occurs unregistering the administration MBean", e);
        }

        synchronized (this) {
            this.notifyAll();
        }
    }

    public boolean isRunning() {
        return this.isRunning;
    }
}
