/**
 * Copyright (c) 2014-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.registry_overlay.core;

import static org.junit.Assert.assertEquals;

import java.io.File;
import java.net.URISyntaxException;
import java.net.URL;
import java.util.Properties;

import org.junit.AfterClass;
import org.junit.BeforeClass;
import org.junit.Test;
import org.ow2.petals.registry_overlay.api.Constants;
import org.ow2.petals.registry_overlay.api.DataModelApi;
import org.ow2.petals.registry_overlay.api.StoredRegistryEndpoint;
import org.ow2.petals.registry_overlay.api.StoredRegistryEndpoint.Type;
import org.ow2.petals.registry_overlay.configuration.generated.PetalsRegistryOverlay;
import org.ow2.petals.registry_overlay.core.exception.ConfigurationException;
import org.ow2.petals.topology.generated.Topology;

import com.ebmwebsourcing.easycommons.lang.reflect.ReflectionHelper;
import com.hazelcast.client.HazelcastClient;
import com.hazelcast.core.HazelcastInstance;

/**
 * Unit tests of {@link ClientListener}
 * 
 * @author Christophe DENEUX - Linagora
 * 
 */
public class ClientListenerTest {

    private static RegistryOverlayNode NODE;

    private static HazelcastInstance HZL_INSTANCE;

    private final static String LOCALHOST = "localhost";

    @BeforeClass
    public static void setupLog() throws URISyntaxException {
        final URL jdkLogCfgFileUrl = Thread.currentThread().getContextClassLoader().getResource("logging.properties");
        System.setProperty("java.util.logging.config.file", new File(jdkLogCfgFileUrl.toURI()).getAbsolutePath());
    }

    @BeforeClass
    public static void startHzlCluster() throws ConfigurationException {
        final PetalsRegistryOverlay petalsRegistryOverlay = ClusterBuilder.createVeryBasicClusterConfig(LOCALHOST);
        final Properties memberProperties = new Properties();
        memberProperties.setProperty(Constants.MEMBER_PROPERTY_NAME_MEMBER_ID, ClusterBuilder.MEMBER_ID);
        NODE = new RegistryOverlayNode(petalsRegistryOverlay, memberProperties);
        NODE.start();

        HZL_INSTANCE = (HazelcastInstance) ReflectionHelper.getPrivateFieldValue(RegistryOverlayNode.class, NODE,
                "instance");
    }

    @AfterClass
    public static void shutdownHzlCluster() {
        NODE.stop();
    }

    /**
     * <p>
     * Check the processing of the event 'CLIENT_DISCONNECTED' on an empty registry.
     * </p>
     * <p>
     * Expected result: no error occur
     * </p>
     */
    @Test
    public void clientDisconnectedOnEmptyRegistry() {

        final HazelcastInstance hzlClient = HazelcastClient.newHazelcastClient(ClusterBuilder
                .createClientConfigForVeryBasicCluster(LOCALHOST));
        hzlClient.shutdown();

    }

    /**
     * <p>
     * Check the processing of the event 'CLIENT_DISCONNECTED' on a registry containing two different standalone
     * topologies.
     * </p>
     * <p>
     * Expected result:
     * <ul>
     * <li>no error occur,</li>
     * <li>on a first client disconnection, the information associated to topology of the client disconnected are
     * cleaned,</li>
     * <li>on a 2nd client disconnection, all information are cleaned,</li>
     * </ul>
     * </p>
     */
    @Test
    public void clientDisconnectedOnRegistryContainingAnotherRegistry() throws InterruptedException {

        final HazelcastInstance hzlClient_1 = HazelcastClient.newHazelcastClient(ClusterBuilder
                .createClientConfigForVeryBasicCluster(LOCALHOST));

        assertEquals(1, HZL_INSTANCE.getClientService().getConnectedClients().size());

        final String domain_1 = "domain_1";
        final String interfaceName_1 = "interface-name-1";
        final String serviceName_1 = "service-name-1";
        final String endpointName_1 = "endpoint-name-1";
        final String registryName_1 = domain_1;

        final StoredRegistryEndpoint endpoint_1 = new StoredRegistryEndpoint(Type.INTERNAL, endpointName_1,
                serviceName_1, new String[] { interfaceName_1 }, "", "");

        try {
            final String clientUUID_1 = hzlClient_1.getLocalEndpoint().getUuid();
            DataModelApi.getClientsMap(hzlClient_1).set(clientUUID_1, registryName_1);
            DataModelApi.getTopologiesMap(hzlClient_1).set(registryName_1, new Topology());
            DataModelApi.getEndpointsMap(hzlClient_1, registryName_1).set(endpointName_1, endpoint_1);
            DataModelApi.getEndpointDescriptionsMap(hzlClient_1, registryName_1).set(endpointName_1,
                    "endpoint-description");

            final HazelcastInstance hzlClient_2 = HazelcastClient.newHazelcastClient(ClusterBuilder
                    .createClientConfigForVeryBasicCluster(LOCALHOST));

            assertEquals(2, HZL_INSTANCE.getClientService().getConnectedClients().size());

            final String domain_2 = "domain_2";
            final String interfaceName_2 = "interface-name-2";
            final String serviceName_2 = "service-name-2";
            final String endpointName_2 = "endpoint-name-2";
            final String registryName_2 = domain_2;

            final StoredRegistryEndpoint endpoint_2 = new StoredRegistryEndpoint(Type.INTERNAL, endpointName_2,
                    serviceName_2, new String[] { interfaceName_2 }, "", "");

            try {
                final String clientUUID_2 = hzlClient_2.getLocalEndpoint().getUuid();
                DataModelApi.getClientsMap(hzlClient_2).set(clientUUID_2, registryName_2);
                DataModelApi.getTopologiesMap(hzlClient_2).set(registryName_2, new Topology());
                DataModelApi.getEndpointsMap(hzlClient_2, registryName_2).set(endpointName_2, endpoint_2);
                DataModelApi.getEndpointDescriptionsMap(hzlClient_2, registryName_2).set(endpointName_2,
                        "endpoint-description");

            } finally {
                hzlClient_2.shutdown();
            }

            // We wait that the event CLIENT_DISCONNECTED is processed
            Thread.sleep(2000);

            assertEquals(1, HZL_INSTANCE.getClientService().getConnectedClients().size());

            assertEquals(0, DataModelApi.getEndpointDescriptionsMap(HZL_INSTANCE, registryName_2).size());
            assertEquals(0, DataModelApi.getEndpointsMap(HZL_INSTANCE, registryName_2).size());

            assertEquals(1, DataModelApi.getEndpointDescriptionsMap(HZL_INSTANCE, registryName_1).size());
            assertEquals(1, DataModelApi.getEndpointsMap(HZL_INSTANCE, registryName_1).size());

            assertEquals(1, DataModelApi.getClientsMap(HZL_INSTANCE).size());
            assertEquals(1, DataModelApi.getTopologiesMap(HZL_INSTANCE).size());
        } finally {
            hzlClient_1.shutdown();
        }

        // We wait that the event CLIENT_DISCONNECTED is processed
        Thread.sleep(2000);

        assertEquals(0, HZL_INSTANCE.getClientService().getConnectedClients().size());

        assertEquals(0, DataModelApi.getEndpointDescriptionsMap(HZL_INSTANCE, registryName_1).size());
        assertEquals(0, DataModelApi.getEndpointsMap(HZL_INSTANCE, registryName_1).size());

        assertEquals(0, DataModelApi.getClientsMap(HZL_INSTANCE).size());
        assertEquals(0, DataModelApi.getTopologiesMap(HZL_INSTANCE).size());
    }

    /**
     * <p>
     * Check the processing of the event 'CLIENT_DISCONNECTED' on a registry containing one topology composed of two
     * containers.
     * </p>
     * <p>
     * Expected result:
     * <ul>
     * <li>no error occur,</li>
     * <li>on a first client disconnection, no information is cleaned,</li>
     * <li>on a 2nd client disconnection, all information are cleaned,</li>
     * </ul>
     * </p>
     */
    @Test
    public void clientDisconnectedOnRegistryContainingOneDistributedRegistry() throws InterruptedException {

        final HazelcastInstance hzlClient_Cont_1 = HazelcastClient.newHazelcastClient(ClusterBuilder
                .createClientConfigForVeryBasicCluster(LOCALHOST));

        assertEquals(1, HZL_INSTANCE.getClientService().getConnectedClients().size());

        final String domain = "my-domain";
        final String interfaceName_1 = "interface-name-1";
        final String serviceName_1 = "service-name-1";
        final String endpointName_1 = "endpoint-name-1";
        final String interfaceName_2 = "interface-name-2";
        final String serviceName_2 = "service-name-2";
        final String endpointName_2 = "endpoint-name-2";
        final String registryName = domain;

        final StoredRegistryEndpoint endpoint_1 = new StoredRegistryEndpoint(Type.INTERNAL, endpointName_1,
                serviceName_1, new String[] { interfaceName_1 }, "", "");

        final StoredRegistryEndpoint endpoint_2 = new StoredRegistryEndpoint(Type.INTERNAL, endpointName_2,
                serviceName_2, new String[] { interfaceName_2 }, "", "");

        try {
            final String clientUUID_1 = hzlClient_Cont_1.getLocalEndpoint().getUuid();
            DataModelApi.getClientsMap(hzlClient_Cont_1).set(clientUUID_1, registryName);
            DataModelApi.getTopologiesMap(hzlClient_Cont_1).set(registryName, new Topology());
            DataModelApi.getEndpointsMap(hzlClient_Cont_1, registryName).set(endpointName_1, endpoint_1);
            DataModelApi.getEndpointDescriptionsMap(hzlClient_Cont_1, registryName).set(endpointName_1,
                    "endpoint-description");

            final HazelcastInstance hzlClient_Cont_2 = HazelcastClient.newHazelcastClient(ClusterBuilder
                    .createClientConfigForVeryBasicCluster(LOCALHOST));

            assertEquals(2, HZL_INSTANCE.getClientService().getConnectedClients().size());

            try {
                final String clientUUID_2 = hzlClient_Cont_2.getLocalEndpoint().getUuid();
                DataModelApi.getClientsMap(hzlClient_Cont_2).set(clientUUID_2, registryName);
                DataModelApi.getTopologiesMap(hzlClient_Cont_2).set(registryName, new Topology());
                DataModelApi.getEndpointsMap(hzlClient_Cont_2, registryName).set(endpointName_2, endpoint_2);
                DataModelApi.getEndpointDescriptionsMap(hzlClient_Cont_2, registryName).set(endpointName_2,
                        "endpoint-description");

            } finally {
                hzlClient_Cont_2.shutdown();
            }

            // We wait that the event CLIENT_DISCONNECTED is processed
            Thread.sleep(2000);

            assertEquals(1, HZL_INSTANCE.getClientService().getConnectedClients().size());

            assertEquals(2, DataModelApi.getEndpointDescriptionsMap(HZL_INSTANCE, registryName).size());
            assertEquals(2, DataModelApi.getEndpointsMap(HZL_INSTANCE, registryName).size());
            assertEquals(1, DataModelApi.getTopologiesMap(HZL_INSTANCE).size());

            assertEquals(1, DataModelApi.getClientsMap(HZL_INSTANCE).size());
        } finally {
            hzlClient_Cont_1.shutdown();
        }

        // We wait that the event CLIENT_DISCONNECTED is processed
        Thread.sleep(2000);

        assertEquals(0, HZL_INSTANCE.getClientService().getConnectedClients().size());

        assertEquals(0, DataModelApi.getEndpointDescriptionsMap(HZL_INSTANCE, registryName).size());
        assertEquals(0, DataModelApi.getEndpointsMap(HZL_INSTANCE, registryName).size());
        assertEquals(0, DataModelApi.getTopologiesMap(HZL_INSTANCE).size());

        assertEquals(0, DataModelApi.getClientsMap(HZL_INSTANCE).size());
    }
}
